/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.geronimo.mail.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;

public class Base64
{
    private static final Encoder encoder = new Base64Encoder();

    /**
     * encode the input data producing a base 64 encoded byte array.
     *
     * @return a byte array containing the base 64 encoded data.
     */
    public static byte[] encode(
        final byte[]    data)
    {
        // just forward to the general array encoder. 
        return encode(data, 0, data.length); 
    }

    /**
     * encode the input data producing a base 64 encoded byte array.
     * 
     * @param data   The data array to encode.
     * @param offset The starting offset within the data array.
     * @param length The length of the data to encode.
     * 
     * @return a byte array containing the base 64 encoded data.
     */
    public static byte[] encode(
        final byte[]    data,
        final int       offset, 
        final int       length)
    {
        final ByteArrayOutputStream    bOut = new ByteArrayOutputStream();

        try
        {
            encoder.encode(data, 0, data.length, bOut);
        }
        catch (final IOException e)
        {
            throw new RuntimeException("exception encoding base64 string: " + e);
        }

        return bOut.toByteArray();
    }

    /**
     * Encode the byte data to base 64 writing it to the given output stream.
     *
     * @return the number of bytes produced.
     */
    public static int encode(
        final byte[]                data,
        final OutputStream    out)
        throws IOException
    {
        return encoder.encode(data, 0, data.length, out);
    }

    /**
     * Encode the byte data to base 64 writing it to the given output stream.
     *
     * @return the number of bytes produced.
     */
    public static int encode(
        final byte[]                data,
        final int                    off,
        final int                    length,
        final OutputStream    out)
        throws IOException
    {
        return encoder.encode(data, off, length, out);
    }

    /**
     * decode the base 64 encoded input data. It is assumed the input data is valid.
     *
     * @return a byte array representing the decoded data.
     */
    public static byte[] decode(
        final byte[]    data)
    {
        // just decode the entire array of data. 
        return decode(data, 0, data.length); 
    }
    

    /**
     * decode the base 64 encoded input data. It is assumed the input data is valid.
     * 
     * @param data   The data array to decode.
     * @param offset The offset of the data array.
     * @param length The length of data to decode.
     * 
     * @return a byte array representing the decoded data.
     */
    public static byte[] decode(
        final byte[]    data, 
        final int       offset, 
        final int       length)
    {
        final ByteArrayOutputStream    bOut = new ByteArrayOutputStream();

        try
        {
            encoder.decode(data, offset, length, bOut);
        }
        catch (final IOException e)
        {
            throw new RuntimeException("exception decoding base64 string: " + e);
        }

        return bOut.toByteArray();
    }

    /**
     * decode the base 64 encoded String data - whitespace will be ignored.
     *
     * @return a byte array representing the decoded data.
     */
    public static byte[] decode(
        final String    data)
    {
        final ByteArrayOutputStream    bOut = new ByteArrayOutputStream();

        try
        {
            encoder.decode(data, bOut);
        }
        catch (final IOException e)
        {
            throw new RuntimeException("exception decoding base64 string: " + e);
        }

        return bOut.toByteArray();
    }

    /**
     * decode the base 64 encoded String data writing it to the given output stream,
     * whitespace characters will be ignored.
     *
     * @return the number of bytes produced.
     */
    public static int decode(
        final String                data,
        final OutputStream    out)
        throws IOException
    {
        return encoder.decode(data, out);
    }

    /**
     * decode the base 64 encoded String data writing it to the given output stream,
     * whitespace characters will be ignored.
     *
     * @param data   The array data to decode.
     * @param out    The output stream for the data.
     *
     * @return the number of bytes produced.
     * @exception IOException
     */
    public static int decode(final byte [] data, final OutputStream out) throws IOException
    {
        return encoder.decode(data, 0, data.length, out);
    }
}
