/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.geronimo.mail.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;

public class UUEncode {
    private static final Encoder encoder = new UUEncoder();

    /**
     * encode the input data producing a UUEncoded byte array.
     *
     * @return a byte array containing the UUEncoded data.
     */
    public static byte[] encode(
        final byte[]    data)
    {
        return encode(data, 0, data.length);
    }

    /**
     * encode the input data producing a UUEncoded byte array.
     *
     * @return a byte array containing the UUEncoded data.
     */
    public static byte[] encode(
        final byte[]    data,
        final int       off,
        final int       length)
    {
        final ByteArrayOutputStream    bOut = new ByteArrayOutputStream();

        try
        {
            encoder.encode(data, off, length, bOut);
        }
        catch (final IOException e)
        {
            throw new RuntimeException("exception encoding UUEncoded string: " + e);
        }

        return bOut.toByteArray();
    }

    /**
     * UUEncode the byte data writing it to the given output stream.
     *
     * @return the number of bytes produced.
     */
    public static int encode(
        final byte[]         data,
        final OutputStream   out)
        throws IOException
    {
        return encoder.encode(data, 0, data.length, out);
    }

    /**
     * UUEncode the byte data writing it to the given output stream.
     *
     * @return the number of bytes produced.
     */
    public static int encode(
        final byte[]         data,
        final int            off,
        final int            length,
        final OutputStream   out)
        throws IOException
    {
        return encoder.encode(data, 0, data.length, out);
    }

    /**
     * decode the UUEncoded input data. It is assumed the input data is valid.
     *
     * @return a byte array representing the decoded data.
     */
    public static byte[] decode(
        final byte[]    data)
    {
        final ByteArrayOutputStream    bOut = new ByteArrayOutputStream();

        try
        {
            encoder.decode(data, 0, data.length, bOut);
        }
        catch (final IOException e)
        {
            throw new RuntimeException("exception decoding UUEncoded string: " + e);
        }

        return bOut.toByteArray();
    }

    /**
     * decode the UUEncided String data.
     *
     * @return a byte array representing the decoded data.
     */
    public static byte[] decode(
        final String    data)
    {
        final ByteArrayOutputStream    bOut = new ByteArrayOutputStream();

        try
        {
            encoder.decode(data, bOut);
        }
        catch (final IOException e)
        {
            throw new RuntimeException("exception decoding UUEncoded string: " + e);
        }

        return bOut.toByteArray();
    }

    /**
     * decode the UUEncoded encoded String data writing it to the given output stream.
     *
     * @return the number of bytes produced.
     */
    public static int decode(
        final String          data,
        final OutputStream    out)
        throws IOException
    {
        return encoder.decode(data, out);
    }
}

